# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

import uuid
import hashlib
import base64
import logging
import json
import http.client
import ssl
import time
from urllib.parse import quote

from typing import Tuple
from aiohttp import web

from foglamp.common import logger
from foglamp.services.management.common import utils
from foglamp.services.management import db
from foglamp.services.core.scheduler.entities import Schedule

__author__ = "Ashish Jabble"
__copyright__ = "Copyright (c) 2020 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_logger = logger.setup(__name__, level=logging.INFO)


INBUILT_FL_SERVICES = ('storage', 'south', 'north')
INBUILT_FL_PLUGINS = ('pi_server', 'ocs', 'OMF')
EXCLUDE_IN_BUILT_SOFTWARES = ["foglamp-north-ocs", "foglamp-north-OMF", "foglamp-north-pi_server",
                              "foglamp-service-south", "foglamp-service-storage", "foglamp-service-north"]


def setup(app: web) -> None:
    app.router.add_route('PUT', '/poll', poll)
    # enable cors support
    utils.enable_cors(app)


def _make_https_request(host: str, action: str, uri: str, payload: dict, port: int = 5000) -> Tuple[int, dict]:
    # FIXME: Not sure how to resolve this circular dep b/w defs
    from foglamp.services.management.server import Server
    header = {'Content-type': 'application/json', 'Authorization': "Bearer {}".format(Server._access_token)}
    data = json.dumps(payload) if payload else None
    conn = http.client.HTTPSConnection(host, port, context=ssl._create_unverified_context())
    conn.request(action, uri, body=data, headers=header)
    r = conn.getresponse()
    if r.status in range(400, 500):
        _logger.error("Could not {}: {}, Client error code: {}, Reason: {}".format(action, uri, r.status, r.reason))
    if r.status in range(500, 600):
        _logger.error("Could not {}: {}, Server error code: {}, Reason: {}".format(action, uri, r.status, r.reason))
    res = r.read().decode()
    response = json.loads(res)
    return r.status, response


def _update_process(db_conn: db, fogman_host: str, current_version: str, fl_name: str) -> None:
    def delete_tables() -> None:
        _tables = ["software", "category", "category_children", "schedule", "filter", "pipeline"]
        for tbl in _tables:
            db.delete_from_table(db_conn, "DELETE FROM {}".format(tbl))

    def insert_schedule_entries(data: dict) -> None:
        sch_list = []
        for sch in data:
            exclusive = 't' if sch['exclusive'] else 'f'
            enabled = 't' if sch['enabled'] else 'f'
            _type = Schedule.Type[sch['type']].value
            s_tuple = (sch['processName'], sch['name'], _type, sch['repeat'], sch['time'], sch['day'],
                       exclusive, enabled)
            sch_list.append(s_tuple)
        sch_query = "INSERT INTO schedule VALUES (?, ?, ?, ?, ?, ?, ?, ?);"
        db.insert_table(db_conn, sch_query, sch_list)

    def insert_category_entries(data: dict) -> None:
        cat_list = []
        for cat in data:
            # FIXME: It should be fix at FogMAN API side, value part should always be return for the category object
            value = cat['value'] if 'value' in cat else {}
            c_tuple = (cat['key'], cat['display_name'], cat['description'], '{}'.format(json.dumps(value)))
            cat_list.append(c_tuple)
        cat_query = "INSERT INTO category VALUES (?, ?, ?, ?);"
        db.insert_table(db_conn, cat_query, cat_list)

    def insert_category_children_entries(data: dict) -> None:
        cat_children_list = []
        for cat in data:
            for child in cat['children']:
                c_tuple = (cat['key'], child)
                cat_children_list.append(c_tuple)
        cat_child_query = "INSERT INTO category_children VALUES (?, ?);"
        db.insert_table(db_conn, cat_child_query, cat_children_list)

    def insert_software_entries(data: dict) -> None:
        software_list = []
        # TODO: Ideally below list should restrict in FogMan API itself
        # or add optional query parameter (skip_in_built_softwares=True/False) on FogLamp API for both plugins/services
        for software in data:
            if software['component'] not in EXCLUDE_IN_BUILT_SOFTWARES:
                s_tuple = (software['component'], software['componentVersion'])
                software_list.append(s_tuple)
        sw_query = "INSERT INTO software VALUES (?, ?);"
        db.insert_table(db_conn, sw_query, software_list)

    def insert_filter_entries(data: dict) -> None:
        filter_list = []
        for _filter in data:
            for k, v in _filter.items():
                f_tuple = (k, json.dumps(v))
                filter_list.append(f_tuple)
        filter_query = "INSERT INTO filter VALUES (?, ?);"
        db.insert_table(db_conn, filter_query, filter_list)

    def insert_pipeline_entries(data: dict) -> None:
        pipeline_list = []
        for k, v in data.items():
            p_tuple = (k, json.dumps(v))
            pipeline_list.append(p_tuple)
        pipeline_query = "INSERT INTO pipeline VALUES (?, ?);"
        db.insert_table(db_conn, pipeline_query, pipeline_list)

    # Fetch data (categories, schedules, softwares) from FogMan
    jdoc = _make_https_request(fogman_host, 'GET', '/manage/version/{}/foglamp/{}/data'.format(
        current_version, fl_name), payload={})
    if jdoc[0] == 200:
        with db_conn:
            # delete all entries from software, schedule, category & category_children tables
            delete_tables()
            # insert new entries
            insert_software_entries(jdoc[1]['data']['software'])
            insert_schedule_entries(jdoc[1]['data']['schedules'])
            insert_category_entries(jdoc[1]['data']['categories'])
            insert_category_children_entries(jdoc[1]['data']['categories'])
            if 'filters' in jdoc[1]['data']:
                insert_filter_entries(jdoc[1]['data']['filters'])
            if 'pipelines' in jdoc[1]['data']:
                insert_pipeline_entries(jdoc[1]['data']['pipelines'])
            # update cached version in instance_version table
            query = "UPDATE instance_version SET cached = '{}'".format(current_version)
            db.update_table(db_conn, query)
    else:
        _logger.warning("Failed to fetch poll data")


def _poll_data(fogman_host: str, fl_instance_name: str) -> str:
    def send_statistics_and_services_to_fogman():
        from foglamp.services.management.server import Server
        ping_info = Server.make_http_request('GET', '/foglamp/ping', payload=None, port=Server._core_service_port)
        data = {"statistics": {"uptime": ping_info["uptime"],
                               "ingress": ping_info["dataRead"],
                               "egress": ping_info["dataSent"],
                               "purged": ping_info["dataPurged"],
                               "health": ping_info["health"]}}
        services = Server.make_http_request('GET', '/foglamp/service', payload=None, port=Server._core_service_port)
        svc_list = []
        for svc in services["services"]:
            svc_list.append({"name": svc["name"], "status": svc["status"]})
        data.update({"services": svc_list})
        _logger.debug('Req: /manage/version/{}/foglamp/{}/poll with data: {}'.format(current_version,
                                                                                     fl_instance_name, data))
        poll_jdoc = _make_https_request(fogman_host, 'PUT', '/manage/version/{}/foglamp/{}/poll'.format(
            current_version, fl_instance_name), payload=data)
        _logger.debug("PUT POLL JDOC---{}".format(poll_jdoc))
        if poll_jdoc[0] == 200:
            _logger.info("Statistics and services details pushed to FogMan successfully.")
        else:
            _logger.error("Failed to push statistics and services details to FogMan!")

    version_jdoc = _make_https_request(fogman_host, 'GET', '/manage/version', payload={})
    version = [v['version'] for v in version_jdoc[1]['versions'] if v['current']]
    current_version = version[0]
    # Always send statistics and services details to FogMan And will be called when manage task schedule triggers
    send_statistics_and_services_to_fogman()
    db_conn = db.create_connection(db._DB_PATH)
    with db_conn:
        row = db.select_table(db_conn, "SELECT * FROM instance_version")
        if not row:
            # initial cached_version is empty
            query = "INSERT INTO instance_version (fogman_host, current, cached, is_synched) VALUES ('{}', " \
                    "'{}', '', 'f')".format(fogman_host, current_version)
            db.insert_table(db_conn, query)
            _update_process(db_conn, fogman_host, current_version, fl_instance_name)
            try:
                _synch_cached_to_foglamp(db_conn)
                message = "Sync complete!"
            except Exception as ex:
                _logger.error("Sync to local FogLAMP failed; {}".format(str(ex)))
                message = "Sync to local FogLAMP is pending."
        else:
            cached_version = row[0][2]
            synched = row[0][3]
            if current_version == cached_version:
                if synched == 't':
                    message = "Current version with FogMan and synchronized cached version are the same. " \
                              "Hence no action required."
                    _logger.warning(message)
                else:
                    msg = "Current version with FogMan and cached version are the same, " \
                          "but not synchronized within FogLAMP..."
                    _logger.warning(msg)
                    try:
                        _synch_cached_to_foglamp(db_conn)
                        message = "Sync complete!"
                    except Exception as ex:
                        _logger.error("Sync to local FogLAMP failed; {}".format(str(ex)))
                        message = "Sync to local FogLAMP is pending."
            else:
                query = "UPDATE instance_version SET current = '{}', is_synched = 'f'".format(current_version)
                db.update_table(db_conn, query)
                _update_process(db_conn, fogman_host, current_version, fl_instance_name)
                try:
                    _synch_cached_to_foglamp(db_conn)
                    message = "Sync complete!"
                except Exception as ex:
                    _logger.error("Sync to local FogLAMP failed; {}".format(str(ex)))
                    message = "Sync to local FogLAMP is pending."
    return message


def _add_fl_and_push_data_to_fogman(fogman_host: str, fl_service_name: str) -> bool:
    # FIXME: Not sure how to resolve this circular dep b/w defs
    from foglamp.services.management.server import Server

    is_pushed = False
    version_jdoc = _make_https_request(fogman_host, 'GET', '/manage/version', payload={})
    # Find unlocked versions and if multiple found then sorting in alphabetical/reverse order.
    # if NOT found version list sorted by created date & in descending order and then create new version with
    # incremented value to the current version of last digit and parent_version is the current version
    # For example: current_version = 1.2.0 then new_version = 1.2.1 & parent_version = 1.2.0
    # NOTE:  Agent schedule task will try to add to new version;
    #        if FL is unknown in any existing current versions. Hence
    #        if the version in which FL was added gets locked but Not set to current i.e deployment is in progress
    versions = [v['version'] for v in version_jdoc[1]['versions'] if not v['locked']]
    if versions:
        versions.sort(reverse=True)
        version = versions[0]
    else:
        ver = version_jdoc[1]['versions']
        # Version list sorted by created date & in descending order
        sorted_version = sorted(ver, key=lambda i: i['created'], reverse=True)
        for v in sorted_version:
            if v['current']:
                base_version = v['version']
                break
            elif v['locked']:
                base_version = v['version']
            else:
                _logger.exception("Failed to find the base version to create new version")
                raise
        major, minor, patch = base_version.split('.')
        version = major + '.' + minor + '.' + str(int(patch) + 1)
        data = {"version": version, "description": "Version via management service"}
        create_version_jdoc = _make_https_request(fogman_host, 'POST', '/manage/version/{}'.format(base_version),
                                                  payload=data)
        _logger.debug("CREATE VERSION JDOC---{}".format(create_version_jdoc))

    data = {"name": fl_service_name, "address": Server._host, "servicePort": Server._core_service_port,
            "username": "admin", "password": "", "certificate": "", "deployMode": "poll"}
    add_manual_fl_jdoc = _make_https_request(fogman_host, 'POST', '/manage/version/{}/foglamp'.format(version),
                                             payload=data)
    _logger.debug("Add FL Manual JDOC---{}".format(add_manual_fl_jdoc))

    uri = "/manage/version/{}/foglamp/{}".format(version, fl_service_name)

    def add_software():
        # Fetch FL installed plugins
        jdoc = Server.make_http_request('GET', '/foglamp/plugins/installed', payload=None,
                                        port=Server._core_service_port)
        fl_plugins = [(p['packageName'], p['version']) for p in jdoc["plugins"] if p['name'] not in INBUILT_FL_PLUGINS]
        _logger.debug("FL Plugins--{}".format(fl_plugins))

        # Fetch FL installed services
        fl_services = Server.make_http_request('GET', '/foglamp/service/installed', payload=None,
                                               port=Server._core_service_port)
        # Version always empty for FL services
        fl_services = [('foglamp-service-{}'.format(s), '') for s in fl_services["services"]
                       if s not in INBUILT_FL_SERVICES]
        _logger.debug("FL Services--{}".format(fl_services))

        # FL softwares = FL Plugins + FL Services
        fl_softwares = fl_plugins
        fl_softwares.extend(fl_services)
        for software in fl_softwares:
            _data = {"component": software[0], "componentVersion": software[1]}
            add_software_jdoc = _make_https_request(fogman_host, 'POST', '{}/software'.format(uri), payload=_data)
            _logger.debug("Add {} Software; JDOC ---{}".format(software, add_software_jdoc))

    def add_category_and_children():
        get_fl_categories = Server.make_http_request('GET', '/foglamp/category', payload=None,
                                                     port=Server._core_service_port)
        for cat in get_fl_categories["categories"]:
            cat_name = quote(cat['key'])
            # Fetch FL Category value by key
            value = Server.make_http_request('GET', '/foglamp/category/{}'.format(cat_name), payload=None,
                                             port=Server._core_service_port)
            _data = {"key": cat['key'],
                     "description": cat['description'],
                     "displayName": cat['displayName'],
                     "value": value
                     }
            add_category_jdoc = _make_https_request(fogman_host, 'POST', '{}/category'.format(uri), payload=_data)
            _logger.debug("Add {} Category; JDOC ---{}".format(cat_name, add_category_jdoc))

        # Add category children
        for cat in get_fl_categories["categories"]:
            cat_name = quote(cat['key'])
            # Fetch FL Category Children
            get_fl_children = Server.make_http_request('GET', '/foglamp/category/{}/children'.format(cat_name),
                                                       payload=None, port=Server._core_service_port)
            children = get_fl_children['categories']
            if children:
                children_list = []
                for child in children:
                    children_list.append(child['key'])
                add_category_children_jdoc = _make_https_request(fogman_host, 'POST', '{}/category/{}/children'.format(
                    uri, cat_name), payload=children_list)
                _logger.debug("Add {} Category Children; JDOC ---{}".format(cat_name, add_category_children_jdoc))

    def add_schedule():
        # Fetch FL Schedules
        get_fl_schedules = Server.make_http_request('GET', '/foglamp/schedule', payload=None,
                                                    port=Server._core_service_port)
        fl_sch = get_fl_schedules["schedules"]
        _logger.debug("FL Schedules: {}".format(fl_sch))
        for sch in fl_sch:
            sch["type"] = Schedule.Type[sch["type"]].name
            sch["day"] = 0 if sch["day"] is None else sch["day"]
            sch["repeat"] = int(sch["repeat"])
            add_schedule_jdoc = _make_https_request(fogman_host, 'POST', '{}/schedule'.format(uri), payload=sch)
            _logger.debug("Add {} schedule; JDOC ---{}".format(sch['name'], add_schedule_jdoc))

    add_software()
    add_category_and_children()
    add_schedule()

    # update state to Managed for instance added in POLL mode after initial sync is completed at FogMan
    data = {"syncStatus": "complete"}
    _logger.debug('/manage/version/{}/foglamp/{}/poll'.format(version, fl_service_name))
    put_state_jdoc = _make_https_request(fogman_host, 'PUT', '/manage/version/{}/foglamp/{}/poll'.format(
        version, fl_service_name), payload=data)
    _logger.debug("PUT FL update state JDOC---{}".format(put_state_jdoc))

    is_pushed = True

    return is_pushed


def _synch_cached_to_foglamp(db_conn: db):
    # FIXME: Not sure how to resolve this circular dep b/w defs
    from foglamp.services.management.server import Server
    roll_back_items = []
    """ Roll back item list will be used if any API call is failed over FL instance """
    MAX_TRIES = 10
    """ Max try count to hit GET package status """
    WAIT_TIME = 15
    """ Sleep/wait time on every GET package status retry """

    # Fetch categories & schedules from FL API
    get_fl_categories = Server.make_http_request('GET', '/foglamp/category', payload=None,
                                                 port=Server._core_service_port)
    get_fl_schedules = Server.make_http_request('GET', '/foglamp/schedule', payload=None,
                                                port=Server._core_service_port)

    def get_diff(old: dict, new: dict) -> list:
        diff = list()
        for key in new:
            if key in old:
                if old[key] != new[key]:
                    diff.append(key)
            else:
                diff.append(key)
        return diff

    def fl_categories() -> list:
        cat_list = []
        for cat in get_fl_categories["categories"]:
            cat_name = quote(cat['key'])
            # Fetch Category value by name from FL API; as its needed to find out the diff with cached schedules
            # And required only for Update action on FL side
            value = Server.make_http_request('GET', '/foglamp/category/{}'.format(cat_name), payload=None,
                                             port=Server._core_service_port)
            fl_temp_dict = {"key": cat['key'],
                            "description": cat['description'],
                            "displayName": cat['displayName'],
                            "value": value
                            }
            cat_list.append(fl_temp_dict)
        return cat_list

    def cached_schedules() -> list:
        # Fetch Cached Schedules
        get_cached_schedules = db.select_table(db_conn, "SELECT * FROM schedule")
        sch_list = []
        # Conversions required for cached schedules; to get actual diff with fl schedules (API vs DB result)
        for sch in get_cached_schedules:
            temp_dict = {
                "processName": sch[0],
                "name": sch[1],
                "type": Schedule.Type(sch[2]).name,
                "repeat": sch[3],
                "time": sch[4],
                "day": sch[5] if sch[5] else None,
                "exclusive": True if sch[6] == 't' else False,
                "enabled": True if sch[7] == 't' else False
            }
            sch_list.append(temp_dict)
        return sch_list

    def cached_categories() -> list:
        # Fetch Cached Categories
        get_cached_categories = db.select_table(db_conn, "SELECT * FROM category")
        cat_list = []
        for cat in get_cached_categories:
            temp_dict = {"key": cat[0],
                         "description": cat[1],
                         "displayName": cat[2],
                         "value": cat[3]
                         }
            cat_list.append(temp_dict)
        return cat_list
    fl_cat = fl_categories()
    _logger.debug("FL Categories: {}".format(fl_cat))
    fl_sch = get_fl_schedules["schedules"]
    _logger.debug("FL Schedules: {}".format(fl_sch))
    cached_cat = cached_categories()
    _logger.debug("Cached Categories: {}".format(cached_cat))
    cached_sch = cached_schedules()
    _logger.debug("Cached Schedules: {}".format(cached_sch))
    get_cached_filters = db.select_table(db_conn, "SELECT * FROM filter")
    _logger.debug("Cached Filters: {}".format(get_cached_filters))
    get_cached_pipelines = db.select_table(db_conn, "SELECT * FROM pipeline")
    _logger.debug("Cached Pipelines: {}".format(get_cached_pipelines))

    def install_software():
        softwares = db.select_table(db_conn, "SELECT * FROM software")

        # Fetch Services from FL API
        fl_services = Server.make_http_request('GET', '/foglamp/service/installed', payload=None,
                                               port=Server._core_service_port)
        fl_services = ['foglamp-service-{}'.format(s) for s in fl_services["services"] if s not in INBUILT_FL_SERVICES]
        _logger.debug("FL Services--{}".format(fl_services))
        # Fetch Cached Services
        cached_services = [(n, v) for n, v in softwares if
                           n not in fl_services and str(n).startswith('foglamp-service-')]
        _logger.debug("Cached Services--{}".format(cached_services))

        svc_payloads = [{"format": "repository", "name": n, "version": v} for n, v in cached_services]
        for data in svc_payloads:
            jdoc = Server.make_http_request('POST', '/foglamp/service?action=install',
                                            payload=data, port=Server._core_service_port)
            _logger.debug("JDOC - POST-SERVICE INSTALLATION: {}".format(jdoc))
            max_retry_count = MAX_TRIES
            while max_retry_count:
                # GET Package Status
                package_status_jdoc = Server.make_http_request('GET', '/{}'.format(jdoc["statusLink"]), payload=data,
                                                               port=Server._core_service_port)
                _logger.debug("JDOC - GET-SERVICE-PACKAGE STATUS: {} on retry count {}".format(package_status_jdoc,
                                                                                               max_retry_count))
                if package_status_jdoc['packageStatus'][0]['status'] == "success":
                    break
                elif package_status_jdoc['packageStatus'][0]['status'] == "failed":
                    # TODO: FOGL-4503 (Low priority) for Software Rollback
                    # And even not required as it is unnecessary overhead
                    raise Exception('{} package installation failed'.format(data['name']))
                # sleep time added b/w retries
                time.sleep(WAIT_TIME)
                max_retry_count -= 1

        # Fetch Plugins from FL API
        jdoc = Server.make_http_request('GET', '/foglamp/plugins/installed', payload=None,
                                        port=Server._core_service_port)
        fl_plugins = [p['packageName'] for p in jdoc["plugins"] if p['name'] not in INBUILT_FL_PLUGINS]
        _logger.debug("FL Plugins--{}".format(fl_plugins))

        prefix_plugins = ['foglamp-south-', 'foglamp-north-', 'foglamp-filter-', 'foglamp-notify-', 'foglamp-rule-']
        # Fetch Cached Plugins
        cached_plugins = [(n, v) for n, v in softwares if n not in fl_plugins and str(n).startswith(
            tuple(prefix_plugins))]
        _logger.debug("Cached Plugins---{}".format(cached_plugins))
        plugin_payloads = [{"format": "repository", "name": n, "version": v} for n, v in cached_plugins]
        for data in plugin_payloads:
            jdoc = Server.make_http_request('POST', '/foglamp/plugins', payload=data, port=Server._core_service_port)
            _logger.debug("JDOC - POST-PLUGINS INSTALL: {}".format(jdoc))
            max_retry_count = MAX_TRIES
            while max_retry_count:
                # GET Package Status
                package_status_jdoc = Server.make_http_request('GET', '/{}'.format(jdoc["statusLink"]), payload=data,
                                                               port=Server._core_service_port)
                _logger.debug("JDOC - GET-PACKAGE STATUS: {} on retry count {}".format(package_status_jdoc,
                                                                                       max_retry_count))
                if package_status_jdoc['packageStatus'][0]['status'] == "success":
                    break
                elif package_status_jdoc['packageStatus'][0]['status'] == "failed":
                    # TODO: FOGL-4503 (Low priority) for Software Rollback
                    # And even not required as it is unnecessary overhead
                    raise Exception('{} package installation failed'.format(data['name']))
                # sleep time added b/w retries
                time.sleep(WAIT_TIME)
                max_retry_count -= 1

    def add_and_update_category():
        # Case of update category:
        for x in fl_cat:
            for y in cached_cat:
                if x['key'] == y['key']:
                    # For matched category; Find diff of cached & FL category and if found then update on FL side
                    fl_cat_list = []
                    cached_cat_list = []
                    for k1, v1 in x['value'].items():
                        fl_cat_list.append({k1: v1['value']})
                    for k2, v2 in json.loads(y['value']).items():
                        cached_cat_list.append({k2: v2['value']})
                    data = {}
                    for config_item_with_value in cached_cat_list:
                        if config_item_with_value not in fl_cat_list:
                            data.update(config_item_with_value)
                    if data:
                        jdoc = Server.make_http_request('PUT', '/foglamp/category/{}'.format(quote(y['key'])),
                                                        payload=data, port=Server._core_service_port)
                        _logger.debug("JDOC - PUT-CATEGORY: {}".format(jdoc))
                        if not isinstance(jdoc, dict):
                            roll_back()
                            raise ValueError('Unable to update Category for name: {} with requested data: {}'.format(
                                y['key'], data))
                        else:
                            # entity, verb, uri, cat name, old data
                            old_data = {}
                            for od in data:
                                _value = [cfg_item[od] for cfg_item in fl_cat_list if od in cfg_item]
                                old_data[od] = _value[0]
                            r_tuple = ("CATEGORY", "PUT", "/foglamp/category", x['key'], old_data)
                            roll_back_items.append(r_tuple)
        # Case of New category
        cached_cat_list = [x['key'] for x in cached_cat]
        fl_cat_list = [y['key'] for y in fl_cat]
        list_diff = list(set(cached_cat_list) - set(fl_cat_list))
        _logger.debug("DIFF of New Category list: {}".format(list_diff))
        for ld in list_diff:
            # Create dict for to accept valid payload format on FL API to create new category
            cat_dict = [x for x in cached_cat if ld == x['key']]
            cat_value = {}
            for k, v in json.loads(cat_dict[0]['value']).items():
                # Copying 'value' to 'default' as POST category to FL for a config item 'value' KV pair is not needed
                if 'default' in v:
                    v['default'] = v['value']
                # Ignoring 'value' KV pair from config item
                if 'value' in v:
                    del v['value']
                cat_value[k] = v
            temp_dict = {"key": cat_dict[0]['key'],
                         "description": cat_dict[0]['description'],
                         "displayName": cat_dict[0]['displayName'],
                         "value": cat_value
                         }
            data = temp_dict
            jdoc = Server.make_http_request('POST', '/foglamp/category', payload=data, port=Server._core_service_port)
            _logger.debug("JDOC - POST-CATEGORY: {}".format(jdoc))
            if not isinstance(jdoc, dict):
                roll_back()
                raise ValueError('Unable to create Category with requested data: {}'.format(data))
            else:
                name = jdoc['key']
                # entity, verb, uri, cat_name, empty data
                r_tuple = ("CATEGORY", "DELETE", "/foglamp/category", name, {})
                roll_back_items.append(r_tuple)
        # Case of New Category Children
        for d in list_diff:
            cat_info = [x for x in cached_cat if d == x['key']]
            # Fetch cached children and if found then create children on FL side
            row = db.select_table(db_conn, "SELECT * FROM category_children where child='{}'".format(
                cat_info[0]['key']))
            if row:
                data = {'children': [row[0][1]]}
                jdoc = Server.make_http_request('POST', '/foglamp/category/{}/children'.format(quote(row[0][0])),
                                                payload=data, port=Server._core_service_port)
                _logger.debug("JDOC - POST-CATEGORY CHILDREN: {}".format(jdoc))
                if not isinstance(jdoc, dict):
                    roll_back()
                    raise ValueError('Unable to create Children for category: {} with payload: {}'.format(
                        cat_info[0]['key'], data))

    def add_and_update_schedule():
        # Case of Update Schedule
        for x in cached_sch:
            for y in fl_sch:
                if x['name'] == y['name']:
                    # For matched schedule; Find diff of cached & FL schedule and if found then update on FL side
                    diff = get_diff(y, x)
                    if diff:
                        data = {}
                        for d in diff:
                            data[d] = Schedule.Type[x[d]].value if 'type' in d else x[d]
                        jdoc = Server.make_http_request('PUT', '/foglamp/schedule/{}'.format(y['id']), payload=data,
                                                        port=Server._core_service_port)
                        _logger.debug("JDOC - PUT-SCHEDULE: {}".format(jdoc))
                        if not isinstance(jdoc, dict):
                            roll_back()
                            raise ValueError('Unable to update Schedule for id: {} with requested data: {}'.format(
                                y['id'], data))
                        else:
                            # entity, verb, uri, sch_id, old data
                            old_data = {}
                            for od in data:
                                if 'type' in od:
                                    _value = Schedule.Type[y[od]].value
                                elif 'repeat' in od:
                                    _value = int(y[od])
                                else:
                                    _value = y[od]
                                old_data[od] = _value
                            r_tuple = ("SCHEDULE", "PUT", "/foglamp/schedule", y['id'], old_data)
                            roll_back_items.append(r_tuple)
        # Case of New Schedule
        cached_sch_list = [x['name'] for x in cached_sch]
        fl_sch_list = [y['name'] for y in fl_sch]
        diff_list = list(set(cached_sch_list) - set(fl_sch_list))
        _logger.debug("DIFF of New Schedule list: {}".format(diff_list))
        for ld in diff_list:
            sch_dict = [sch for sch in cached_sch if ld == sch['name']]
            # Create dict for to accept valid payload format on FL API to create new schedule
            temp = sch_dict[0]
            temp["type"] = Schedule.Type[temp["type"]].value
            temp["day"] = 0 if temp["day"] is None else temp["day"]
            temp["process_name"] = temp["processName"]
            del temp["processName"]
            data = temp
            jdoc = Server.make_http_request('POST', '/foglamp/schedule', payload=data,
                                            port=Server._core_service_port)
            _logger.debug("JDOC - POST-SCHEDULE: {}".format(jdoc))
            if not isinstance(jdoc, dict):
                roll_back()
                raise ValueError('Unable to create Schedule with requested data: {}'.format(data))
            else:
                sch_id = jdoc['schedule']['id']
                # entity, verb, uri, sch_id, empty data
                r_tuple = ("SCHEDULE", "DELETE", "/foglamp/schedule", sch_id, {})
                roll_back_items.append(r_tuple)

    def add_and_update_filters() -> None:
        for c in get_cached_filters:
            try:
                res = next(item for item in get_cached_pipelines if c[0] in json.loads(item[1]))
                svc_filter_cat_name = "{}_{}".format(res[0], c[0])
            except Exception:
                _logger.error("{} not found in pipelines".format(c[0]))
            _logger.debug("Filter name: {}".format(svc_filter_cat_name))
            new_dict = json.loads(c[1])['filter_config']
            old_dict = {}
            try:
                res = next(item for item in fl_cat if item["key"] == svc_filter_cat_name)
                _logger.debug("Filter Category found: {}".format(res))
                for k, v in res['value'].items():
                    if k in new_dict:
                        old_dict[k] = v['value']
                _logger.debug("OLD Filter config: {}".format(old_dict))
                _logger.debug("New Filter config: {}".format(new_dict))
                diff = get_diff(old_dict, new_dict)
                _logger.debug("DIFF: {}".format(diff))
                payload = {}
                for d in diff:
                    payload.update({d: new_dict[d]})
                if payload:
                    bulk_update_jdoc = Server.make_http_request('PUT', '/foglamp/category/{}'.format(
                        quote(svc_filter_cat_name)), payload=payload, port=Server._core_service_port)
                    _logger.debug("JDOC - PUT-CATEGORY: {} with payload: {}".format(bulk_update_jdoc, payload))
            except Exception:
                _logger.debug("Case of POST Filter for: {} with payload {}".format(c[0], c[1]))
                jdoc = Server.make_http_request('POST', '/foglamp/filter', payload=json.loads(c[1]),
                                                port=Server._core_service_port)
                _logger.debug("JDOC - POST-Filter: {}".format(jdoc))

    def add_and_update_pipelines() -> None:
        for _pipeline in get_cached_pipelines:
            get_pipeline_jdoc = Server.make_http_request('GET', '/foglamp/filter/{}/pipeline'.format(
                quote(_pipeline[0])), payload=None, port=Server._core_service_port)
            _logger.debug("JDOC - GET Pipeline for {} service: {}".format(_pipeline[0], get_pipeline_jdoc))
            new_pipeline = {"pipeline": json.loads(_pipeline[1])}
            if isinstance(get_pipeline_jdoc, dict):
                old_pipeline = get_pipeline_jdoc['result']['pipeline']
                pipeline_diff = list(set(old_pipeline) - set(new_pipeline['pipeline']))
                uri = "/foglamp/filter/{}/pipeline?allow_duplicates=true&append_filter=false".format(quote(_pipeline[0]))
                put_pipeline_jdoc = Server.make_http_request('PUT', uri, payload=new_pipeline,
                                                             port=Server._core_service_port)
                _logger.debug("JDOC - PUT Pipeline for {} service: {}".format(_pipeline[0], put_pipeline_jdoc))
                for d in pipeline_diff:
                    delete_cat_jdoc = Server.make_http_request('DELETE', '/foglamp/category/{}'.format(quote(d)),
                                                               payload=None, port=Server._core_service_port)
                    _logger.debug("JDOC - Filter Standalone DELETE CATEGORY: {}".format(delete_cat_jdoc))
            else:
                uri = "/foglamp/filter/{}/pipeline?allow_duplicates=true&append_filter=false".format(
                    quote(_pipeline[0]))
                put_pipeline_jdoc = Server.make_http_request('PUT', uri, payload=new_pipeline,
                                                             port=Server._core_service_port)
                _logger.debug("JDOC - PUT Pipeline for {} service: {}".format(_pipeline[0], put_pipeline_jdoc))

    def delete_entries_from_fl() -> None:
        fl_cat_list = [y1['key'] for y1 in fl_cat]
        cached_cat_list = [x1['key'] for x1 in cached_cat]
        diff_list_cat = list(set(fl_cat_list) - set(cached_cat_list))
        delete_cat_list = []
        for delete_cat in diff_list_cat:
            if not (delete_cat.endswith("Advanced")):
                try:
                    sch_id = next(item['id'] for item in fl_sch if
                                  item["name"] == delete_cat and item["processName"] in ['south_c', 'north_C'])
                    if sch_id:
                        delete_svc_jdoc = Server.make_http_request('DELETE', '/foglamp/service/{}'.format(
                            quote(delete_cat)), payload=None, port=Server._core_service_port)
                        _logger.debug("JDOC - DELETE SERVICE: {}".format(delete_svc_jdoc))
                        # Extra wait time needed to complete recursively deletion
                        time.sleep(WAIT_TIME - 10)
                except Exception:
                    try:
                        sch_id = next(item['id'] for item in fl_sch if
                                      item["name"] == delete_cat and item["processName"] == 'north_c')
                        if sch_id:
                            delete_task_jdoc = Server.make_http_request('DELETE', '/foglamp/scheduled/task/{}'.format(
                                quote(delete_cat)), payload=None, port=Server._core_service_port)
                            _logger.debug("JDOC - DELETE TASK: {}".format(delete_task_jdoc))
                            # Extra wait time needed to complete recursively deletion
                            time.sleep(WAIT_TIME - 10)
                    except Exception:
                        # Don't delete Parent categories Notification, North, South
                        if delete_cat not in ['Notifications', 'North', 'South']:
                            delete_cat_list.append(delete_cat)
        # Filters
        get_fl_filters = Server.make_http_request('GET', '/foglamp/filter', payload=None,
                                                  port=Server._core_service_port)
        fl_filter = get_fl_filters["filters"]
        fl_filter_list = [y1['name'] for y1 in fl_filter]
        temp = fl_filter_list.copy()
        for f in temp:
            get_fl_filter_user = Server.make_http_request('GET', '/foglamp/filter/{}'.format(quote(f)), payload=None,
                                                          port=Server._core_service_port)
            if isinstance(get_fl_filter_user, dict):
                fl_filter_user = get_fl_filter_user["filter"]["users"]
                for fuser in fl_filter_user:
                    fl_filter_list.append('{}_{}'.format(fuser, f))
                    fl_filter_list.append('{} Filters'.format(fuser))
        _logger.debug("DELETE Filter Name List: {}".format(fl_filter_list))
        _logger.debug("DELETE CAT LIST: {}".format(delete_cat_list))
        others_delete_list = list(set(delete_cat_list) - set(fl_filter_list))
        _logger.debug("DELETE OTHERS LIST: {}".format(others_delete_list))
        for c in others_delete_list:
            delete_cat_jdoc = Server.make_http_request('DELETE', '/foglamp/category/{}'.format(quote(c)),
                                                       payload=None, port=Server._core_service_port)
            _logger.debug("JDOC - DELETE CATEGORY: {}".format(delete_cat_jdoc))
        # Schedules
        cached_sch_list = [x2['name'] for x2 in cached_sch]
        fl_sch_list = [y2['name'] for y2 in fl_sch]
        diff_list_sch = list(set(fl_sch_list) - set(cached_sch_list))
        _logger.debug("DIFF of Delete Schedule list: {}".format(diff_list_sch))
        others_delete_sch_list = list(set(diff_list_sch) - set(diff_list_cat))
        _logger.debug("Others DELETE Schedule LIST: {}".format(others_delete_sch_list))
        for delete_cat in others_delete_sch_list:
            try:
                sch_id = next(item['id'] for item in fl_sch if item["name"] == delete_cat)
                delete_sch_jdoc = Server.make_http_request('DELETE', '/foglamp/schedule/{}'.format(sch_id),
                                                           payload=None, port=Server._core_service_port)
                _logger.debug("JDOC - DELETE SCHEDULE: {}".format(delete_sch_jdoc))
            except Exception:
                pass

    def roll_back() -> None:
        _logger.debug("ROLL BACK ITEMS: {}".format(roll_back_items))
        for item in roll_back_items:
            _logger.debug("ROLL BACK item: {}".format(item))
            entity = item[0]
            verb = item[1]
            uri = item[2]
            if entity == 'CATEGORY':
                cat_name = quote(item[3])
                if verb == 'PUT':
                    data = {}
                    for k, v in item[4].items():
                        data[k] = v
                    if data:
                        jdoc = Server.make_http_request('{}'.format(verb), '{}/{}'.format(uri, cat_name), payload=data,
                                                        port=Server._core_service_port)
                        _logger.debug("JDOC - ROLL BACK PUT-CATEGORY: {}".format(jdoc))
                elif verb == 'DELETE':
                    jdoc = Server.make_http_request('{}'.format(verb), '{}/{}'.format(uri, cat_name), payload={},
                                                    port=Server._core_service_port)
                    _logger.debug("JDOC - ROLL BACK DELETE-CATEGORY: {}".format(jdoc))
                else:
                    _logger.error("For entity {} invalid verb found: {} when rollback".format(entity, verb))
            elif entity == 'SCHEDULE':
                sch_id = item[3]
                if verb == 'PUT':
                    data = {}
                    for k, v in item[4].items():
                        data[k] = v
                    if data:
                        jdoc = Server.make_http_request('{}'.format(verb), '{}/{}'.format(uri, sch_id), payload=data,
                                                        port=Server._core_service_port)
                        _logger.debug("JDOC - ROLL BACK PUT-SCHEDULE: {}".format(jdoc))
                elif verb == 'DELETE':
                    jdoc = Server.make_http_request('{}'.format(verb), '{}/{}'.format(uri, sch_id), payload={},
                                                    port=Server._core_service_port)
                    _logger.debug("JDOC - ROLL BACK DELETE-SCHEDULE: {}".format(jdoc))
                else:
                    _logger.error("For entity {} invalid verb found: {} when rollback".format(entity, verb))
            else:
                _logger.error("For Rollback, bad entity found: {}".format(entity))

    install_software()
    add_and_update_category()
    add_and_update_schedule()
    add_and_update_filters()
    add_and_update_pipelines()
    delete_entries_from_fl()

    # Set is_synched to True for instance_version table
    query = "UPDATE instance_version SET is_synched = 't'"
    db.update_table(db_conn, query)


def get_fogman_oauth2_access_token(fogman_host: str, username: str, password: str):
    def get_oauth2_services():
        oauth2_svc_jdoc = _make_https_request(fogman_host, 'GET', '/oauth2/services', payload={})
        if oauth2_svc_jdoc[0] == 200:
            return oauth2_svc_jdoc[1]['oauth2_services'][0]
        else:
            _logger.error("GET /oauth2/services request is failed. Got {} HTTP status code and reason {}".format(
                oauth2_svc_jdoc[0], oauth2_svc_jdoc[1]))
            raise

    def get_oauth2_authorize():
        _type = 'token'
        _scope = 'all'
        url = "{}?client_id={}&redirect_uri=http%3A%2F%2F{}%3A12345&response_type={}&scope={}&state={}".format(
            authorize_uri, client_id, fogman_host, _type, _scope, base64_oauth2_authorize_state)
        conn = http.client.HTTPSConnection(fogman_host, server_port, context=ssl._create_unverified_context())
        conn.request('GET', url, body=None)
        authorize_response = conn.getresponse()
        if authorize_response.status == 302:
            # Extract header fields info which we need during oAuth2 login in next step
            _location = authorize_response.getheader('Location')
            _cookie = authorize_response.getheader('Set-Cookie').split(';')[0]
            return _location, _cookie
        else:
            _logger.error("GET /oauth2/authorize request is failed. Got {} HTTP status code and reason {}".format(
                authorize_response.status, authorize_response.reason))
            raise

    def post_login():
        headers = {"Content-Type": "application/x-www-form-urlencoded", "Cookie": "{}".format(cookie)}
        # base64 string for password
        pwd = base64.b64encode(hashlib.sha256(password.encode('utf-8')).digest()).decode("utf-8")
        url = "{}?username={}&password={}".format(location, username, pwd)
        conn = http.client.HTTPSConnection(fogman_host, server_port, context=ssl._create_unverified_context())
        conn.request('POST', url, body=None, headers=headers)
        login_response = conn.getresponse()
        if login_response.status == 302:
            # Access token in the URI fragment after #char
            loc = login_response.getheader('Location').split("#")
            # State in the URI fragment And verify with which we sent during GET /oauth2/authorize
            state = loc[1].split('&state=')[1].split('&')[0]
            if state == base64_oauth2_authorize_state:
                return loc[1].split('&')[0].replace('access_token=', '')
            else:
                _logger.error("State parameter does not match that one sent in /oauth2/authorize request.")
                raise
        else:
            _logger.error("POST /oauth2/login request is failed. Got {} HTTP status code and reason {}".format(
                login_response.status, login_response.reason))
            raise

    # Encoded base64 string for state (i.e MGT-LOGIN_uuid)
    base64_oauth2_authorize_state = base64.b64encode("MGT-LOGIN_{}".format(
        uuid.uuid4()).encode('ascii')).decode('ascii')
    # GET oAuth2 services from FogMan API
    fogman_svc = get_oauth2_services()
    # Below are the values are required to be passed for oAuth2 authorize request
    client_id = fogman_svc['oauth2_app_id']
    server_url = fogman_svc['oauth2_server_name']
    server_port = server_url.split(':')[-1]
    authorize_uri = fogman_svc['oauth2_authorize_uri']
    # oAuth2 Authorize Request
    location, cookie = get_oauth2_authorize()
    # oAuth2 Login Request
    oauth2_token = post_login()
    return oauth2_token


async def poll(request: web.Request) -> web.Response:
    # To avoid circular import
    from foglamp.services.management.server import Server

    if Server._config['enable']['value'] == 'false':
        msg = "operation cannot be executed as service configuration in disabled state."
        _logger.warning(msg)
        return web.json_response({"message": msg})

    fogman_host = Server._config['manager']['value']
    response = {"message": {}}
    try:
        # Get service name of FogLAMP instance
        svc_category_jdoc = Server.make_http_request('GET', '/foglamp/category/service', payload=None,
                                                     port=Server._core_service_port)
        fl_service_name = svc_category_jdoc['name']['value']
        # NOTE:
        # oAuth2 implementation as per Specs - All clients (GUI, CLI, Mgt) will use same fogman CLIENT/APP ID
        # Some time in the future we support using external auth services then there will be more than one.
        # For the time we are safe to assume just a single entry (i.e /oauth2/services ) in the response.
        # Get oAuth2 Access Token

        Server._access_token = get_fogman_oauth2_access_token(fogman_host=fogman_host,
                                                              username=Server._config['username']['value'],
                                                              password=Server._config['password']['value'])
        _logger.debug("Access_token: {}".format(Server._access_token))
        # NOTE:
        # 1) It is expected to register FL instance with the name in service category.
        #    If already known to FogMan, then what name FogLAMP should use (and update service category before add?)
        #  2) knowing by name is enough? what if address mismatch
        #     Let say I had F1 name (and it was you who registred already with F1) ;
        #     checking with name will say oh yeah you are known to me.
        #  Probable solution for #2:
        #   a) Fetch the version /status/foglamp/F1
        #   b) Find the address manage/version/{version}/foglamp
        # Check FL is known to the FogMan
        fl_status_jdoc = _make_https_request(fogman_host, 'GET', '/status/foglamp/{}'.format(fl_service_name),
                                             payload={})
        if fl_status_jdoc[0] == 200:
            deployed_version = fl_status_jdoc[1]['status']['deployed']['version']
            response['message'] = _poll_data(fogman_host, fl_service_name)
        elif fl_status_jdoc[0] == 404:
            # FL node is NOT known to FogMan
            is_pushed = _add_fl_and_push_data_to_fogman(fogman_host, fl_service_name)
            response['message'] = "FL data pushed to FogMan successfully!" if is_pushed \
                else "Failed to Push data to FogMan"
        elif fl_status_jdoc[0] == 400:
            # If no current version found, attached FL instance to default 1.0.0 version and then data push to FogMan
            if fl_status_jdoc[1]['error'] == "Current version not found":
                is_pushed = _add_fl_and_push_data_to_fogman(fogman_host, fl_service_name)
                response['message'] = "FL data pushed to FogMan successfully!" if is_pushed \
                    else "Failed to Push data to FogMan"
            else:
                raise RuntimeError(fl_status_jdoc[1]['error'])
        else:
            raise RuntimeError(fl_status_jdoc[1]['error'])
    except RuntimeError as err:
        response['message'] = str(err)
        _logger.error("Internal Server error: {}".format(response))
        return web.HTTPInternalServerError(reason=response, body=json.dumps(response))
    except Exception as ex:
        response['message'] = "{}. See detailed exception message in syslog.".format(str(ex))
        _logger.exception("Poll operation failed due to reason: {}".format(str(ex)))
        return web.HTTPInternalServerError(reason=str(ex), body=json.dumps(response))
    else:
        return web.json_response(response)
